from __future__ import annotations

import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import timedelta, datetime
from typing import Dict, Optional, Set, List

import pandas as pd

LOCAL_TZ = datetime.now().astimezone().tzinfo
UTC_TZ = "UTC"


class Persistence(ABC):
    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def get_commit(
        self, commit_id: Optional[str] = None, commit_num: Optional[int] = None
    ):
        pass

    @abstractmethod
    def get_latest_commit(self):
        pass

    @abstractmethod
    def commit(
        self,
        commit: Commit,
        dirty_keys: Optional[Set[str]] = None,
    ) -> str:
        pass

    @abstractmethod
    def search_commits(
        self, label: Optional[str] = None, key: Optional[str] = None
    ) -> List[Commit]:
        pass

    @abstractmethod
    def save_temp_commit(self, commit):
        pass

    @abstractmethod
    def load_temp_commit(self):
        pass

    @staticmethod
    def stamp_dirty_params_with_commit(
        commit: Commit, dirty_keys: Optional[Set[str]] = None
    ):
        if dirty_keys:
            for key in dirty_keys:
                if key in commit.params:
                    param = commit.params[key]
                    param.commit_id = commit.id
                    if isinstance(param.expiration, timedelta):
                        param.expiration = commit.timestamp + param.expiration


@dataclass
class Commit:
    params: Dict
    tags: Dict
    id: Optional[str] = None  # commit_id
    timestamp: Optional[pd.Timestamp] = None
    label: Optional[str] = None

    def __post_init__(self):
        self.id = self.id or ""
        self.timestamp = self.timestamp or pd.Timestamp(time.time_ns())
        self.label = self.label or None
        self.params = self.params or {}
        self.tags = self.tags or {}

    def to_metadata(self) -> Metadata:
        return Metadata(self.id, self.timestamp, self.label)

    @property
    def local_timestamp(self):
        return self.timestamp.tz_localize(UTC_TZ).tz_convert(LOCAL_TZ)


@dataclass
class Metadata:
    id: str  # commit id
    timestamp: pd.Timestamp
    label: Optional[str] = None

    def __post_init__(self):
        self.id = self.id or ""
        self.timestamp = self.timestamp or pd.Timestamp(time.time_ns())



"""experiments_favorite_col

Revision ID: 273a9fae6206
Revises: 7fa75ca1263f
Create Date: 2022-06-23 10:16:39.441194+00:00

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "273a9fae6206"
down_revision = "7fa75ca1263f"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "Experiments",
        sa.Column(
            "favorite",
            sa.Boolean(),
            nullable=False,
            default=False,
            server_default=text("False"),
        ),
    )


def downgrade():
    op.drop_column("Experiments", "favorite")



import uuid

from sqlalchemy import Column, String, DateTime
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import declarative_base, declarative_mixin
from sqlalchemy.sql import expression

Base = declarative_base()

""" Auto-generate UTC timestamp in PostgreSQL """


class UtcNow(expression.FunctionElement):
    """SqlAlchemy function to generate UTC timestamp on the server-side (Postgres)
    Source: https://docs.sqlalchemy.org/en/14/core/compiler.html#utc-timestamp-function
    """

    type = DateTime()
    inherit_cache = True


@compiles(UtcNow, "postgresql")
def pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


""" Compiling to sqlite for testing purposes """


@compiles(UtcNow, "sqlite")
def sqlite_utcnow(element, compiler, **kw):
    return "CURRENT_TIMESTAMP"


@compiles(UUID, "sqlite")
def sqlite_uuid(element, compiler, **kw):
    return "TEXT"


@compiles(JSONB, "sqlite")
def sqlite_jsonb(element, compiler, **kw):
    return "JSON"


@compiles(DateTime, "sqlite")
def sqlite_datetime(element, compiler, **kw):
    return "TEXT"


""" SqlAlchemy ORM model for ParamStore """


@declarative_mixin
class CommitMixin:
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    timestamp = Column(
        DateTime(timezone=False), nullable=False, server_default=UtcNow()
    )
    label = Column(String(256))
    params = Column(JSONB, nullable=False)
    tags = Column(JSONB, nullable=False)


class CommitTable(CommitMixin, Base):
    __tablename__ = "commit"
    # id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    # timestamp = Column(
    #     DateTime(timezone=False), nullable=False, server_default=UtcNow()
    # )
    # label = Column(String(256))
    # params = Column(JSONB, nullable=False)
    # tags = Column(JSONB, nullable=False)

    def __repr__(self):
        return f"<CommitTable(commit_id={self.id}, label={self.label})>"


class TempTable(CommitMixin, Base):
    __tablename__ = "temp"

    def __repr__(self):
        return f"<TempTable(commit_id={self.id}, label={self.label})>"



"""wrapping_param_store_values

Revision ID: 06140c96c8c4
Revises: 9ffd2ba0d5bf
Create Date: 2022-05-19 09:12:01.569825+00:00

"""

import shutil

from entropylab.logger import logger
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.params.persistence.migrations import (
    migrate_param_store_0_1_to_0_2,
)
from entropylab.pipeline.params.persistence.tinydb.tinydbpersistence import set_version
from entropylab.pipeline.results_backend.sqlalchemy.alembic.alembic_util import (
    AlembicUtil,
)

# revision identifiers, used by Alembic.
revision = "06140c96c8c4"
down_revision = "9ffd2ba0d5bf"
branch_labels = None
depends_on = None


def upgrade():
    path = str(AlembicUtil.get_param_store_file_path())
    logger.debug(f"Attempting to migrate ParamStore file {path} from v0.1 to v0.2")
    try:
        migrate_param_store_0_1_to_0_2(path, revision)
        set_version(path, "0.2", revision)
    except EntropyError as ee:
        logger.warning(str(ee))
    logger.debug("Done migrating from v0.1 to v0.2")


def downgrade():
    path = AlembicUtil.get_param_store_file_path()
    backup_json_file_path = path.replace(".json", ".json.{revision}.bak")
    downgraded_json_file_path = path.replace(".json", ".json.{revision}.downgraded")
    logger.debug(
        f"Attempting to downgrade ParamStore file {path} from v0.2 "
        f"to v0.1 using backup file {backup_json_file_path}"
    )
    shutil.move(path, downgraded_json_file_path)
    shutil.move(backup_json_file_path, path)
    logger.debug("Done downgrading from v0.2 to v0.1")



from __future__ import annotations

import copy
import os
import shutil
from pathlib import Path

import pandas as pd
from tinydb import TinyDB

from entropylab.pipeline.params.param_store import Param
from entropylab.pipeline.params.persistence.tinydb.storage import JSONPickleStorage
from entropylab.pipeline.params.persistence.tinydb.tinydbpersistence import (
    check_version,
    TEMP_TABLE,
    TEMP_DOC_ID,
    INFO_TABLE,
    INFO_DOC_ID,
)


def fix_param_qualified_name(path: str | Path, revision: str):
    """
    Backup and fix the fully qualified names of Param in an ParamStore JSON
    file.

    :param path: path to an existing JSON TinyDB file containing params.
    :param revision:
    """
    if not os.path.isfile(path):
        return
    old_name = "entropylab.api.in_process_param_store.Param"
    new_name = "entropylab.pipeline.api.param_store.Param"
    path = str(path)
    backup_path = _backup_file(path, revision)
    # TODO: Support locking
    with TinyDB(backup_path) as old_db:
        old_commits = old_db.all()
        old_temp = old_db.table(TEMP_TABLE).get(doc_id=TEMP_DOC_ID)
        old_info = old_db.table(INFO_TABLE).get(doc_id=INFO_DOC_ID)
    with TinyDB(path) as new_db:
        for old_commit in old_commits:
            new_commit = copy.deepcopy(old_commit)
            _rename_param_qualified_names(new_commit, old_name, new_name)
            new_db.insert(new_commit)
        if old_temp:
            new_temp = copy.deepcopy(old_temp)
            _rename_param_qualified_names(new_temp, old_name, new_name)
            new_db.table(TEMP_TABLE).insert(new_temp)
        if old_info:
            new_info = copy.deepcopy(old_info)
            new_db.table(INFO_TABLE).insert(new_info)


def migrate_param_store_0_1_to_0_2(path: str | Path, revision: str) -> None:
    """
    Backup and migrate an ParamStore JSON file from storing values to storing
    Params containing the values. Preserves commits, timestamps and ids.

    :param path: path to an existing JSON TinyDB file containing params.
    :param revision: the Alembic revision (version) that calls this function

    """
    old_version = "0.1"
    new_version = "0.2"

    path = str(path)
    check_version(path, old_version, new_version)
    backup_path = _backup_file(path, revision)
    with TinyDB(backup_path) as old_db:
        old_commits = old_db.all()
        old_temp = old_db.table(TEMP_TABLE).get(doc_id=TEMP_DOC_ID)
    with TinyDB(path, storage=JSONPickleStorage) as new_db:
        for old_commit in old_commits:
            new_commit = copy.deepcopy(old_commit)
            _wrap_params(new_commit)
            _rename_ns(new_commit)
            new_db.insert(new_commit)
        if old_temp:
            new_temp = copy.deepcopy(old_temp)
            _wrap_params(new_temp)
            _rename_ns(new_temp)
            new_db.table(TEMP_TABLE).insert(new_temp)


def _backup_file(path: str, revision: str):
    backup_path = f"{path}.{revision}.bak"
    shutil.move(path, backup_path)
    return backup_path


def _wrap_params(new_commit):
    params = new_commit["params"]
    for key in params.keys():
        value = params[key]
        if not isinstance(value, Param):
            params[key] = Param(value)
    new_commit["params"] = params


def _rename_ns(new_commit):
    timestamp = new_commit["metadata"]["ns"]
    del new_commit["metadata"]["ns"]
    new_commit["metadata"]["timestamp"] = timestamp


def migrate_param_store_0_2_to_0_3(path: str | Path, revision: str) -> None:
    """
    Backup and migrate an ParamStore JSON file to use the newer qualified name for Param
    instances and convert timestamp and expiration values from int to pd.Timestamp.

    :param path: path to an existing JSON TinyDB file containing params.
    :param revision: the Alembic revision (version) that calls this function.

    """
    old_version = "0.2"
    new_version = "0.3"

    old_name = "entropylab.pipeline.api.param_store.Param"
    new_name = ("entropylab.pipeline.params.param_store.Param",)

    path = str(path)
    check_version(path, old_version, new_version)
    backup_path = _backup_file(path, revision)
    with TinyDB(backup_path, storage=JSONPickleStorage) as old_db:
        old_commits = old_db.all()
        old_temp = old_db.table(TEMP_TABLE).get(doc_id=TEMP_DOC_ID)
        old_info = old_db.table(INFO_TABLE).get(doc_id=INFO_DOC_ID)
    with TinyDB(path, storage=JSONPickleStorage) as new_db:
        for old_commit in old_commits:
            new_commit = copy.deepcopy(old_commit)
            _convert_timestamp_to_pd_timestamp(new_commit, old_commit)
            _convert_expiration_to_pd_timestamp(new_commit)
            _rename_param_qualified_names(new_commit, old_name, new_name)
            new_db.insert(new_commit)
        if old_temp:
            new_temp = copy.deepcopy(old_temp)
            _convert_timestamp_to_pd_timestamp(new_temp, old_temp)
            _convert_expiration_to_pd_timestamp(new_temp)
            _rename_param_qualified_names(new_temp, old_name, new_name)
            new_db.table(TEMP_TABLE).insert(new_temp)
        if old_info:
            new_info = copy.deepcopy(old_info)
            new_db.table(INFO_TABLE).insert(new_info)


def _rename_param_qualified_names(new_commit, old_name, new_name):
    for param in new_commit["params"].values():
        if "py/object" in param and param["py/object"] == old_name:
            param["py/object"] = new_name


def _convert_timestamp_to_pd_timestamp(new_commit, old_commit):
    new_commit["metadata"]["timestamp"] = pd.Timestamp(
        old_commit["metadata"]["timestamp"]
    )


def _convert_expiration_to_pd_timestamp(new_commit):
    commit_timestamp = new_commit["metadata"]["timestamp"]
    params = new_commit["params"]
    for _, param in params.items():
        if isinstance(param.expiration, int):
            param.expiration = commit_timestamp + pd.Timedelta(seconds=param.expiration)



import os
import uuid
from uuid import UUID
from pathlib import Path
from typing import Optional, Set, List

import jsonpickle
from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Connection
from sqlalchemy.orm import sessionmaker, close_all_sessions

from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.params.persistence.persistence import Persistence, Commit
from entropylab.pipeline.params.persistence.sqlalchemy.model import (
    CommitTable,
    TempTable,
)

TEMP_COMMIT_ID = UUID("00000000-0000-0000-0000-000000000000")


class SqlAlchemyPersistence(Persistence):
    def __init__(self, url: Optional[str] = None):
        self.engine = create_engine(
            url,
            json_serializer=jsonpickle.encode,
            json_deserializer=jsonpickle.decode,
        )
        self.__session_maker = sessionmaker(bind=self.engine)
        if self._db_is_empty():
            self.__init()

    def __init(self):
        with self.engine.connect() as connection:
            alembic_cfg = self.__alembic_build_config(connection)
            command.upgrade(alembic_cfg, "head")

    def __alembic_build_config(self, connection: Connection) -> Config:
        config_location = self._abs_path_to("alembic.ini")
        script_location = self._abs_path_to("alembic")
        alembic_cfg = Config(config_location)
        alembic_cfg.set_main_option("script_location", script_location)
        """
        Modified by @urig to share a single engine across multiple (typically in-memory)
        connections, based on this cookbook recipe:
        https://alembic.sqlalchemy.org/en/latest/cookbook.html#connection-sharing
        """
        alembic_cfg.set_main_option("sqlalchemy.url", "sqlite:///:memory:")
        alembic_cfg.attributes["connection"] = connection  # overrides dummy url above
        return alembic_cfg

    @staticmethod
    def _abs_path_to(rel_path: str) -> str:
        source_path = Path(__file__).resolve()
        source_dir = source_path.parent
        return os.path.join(source_dir, rel_path)

    def close(self):
        close_all_sessions()

    def get_commit(
        self, commit_id: Optional[str] = None, commit_num: Optional[int] = None
    ):
        if commit_id:
            with self.__session_maker() as session:
                commit = (
                    session.query(CommitTable)
                    .filter(CommitTable.id == UUID(commit_id))
                    .one_or_none()
                )
                if commit:
                    return commit
                else:
                    raise EntropyError(f"Commit with id '{commit_id}' not found")
        elif commit_num:
            with self.__session_maker() as session:
                commit = (
                    session.query(CommitTable)
                    .order_by(CommitTable.timestamp.asc())
                    .offset(commit_num - 1)
                    .limit(1)
                    .one_or_none()
                )
                if commit:
                    return commit
                else:
                    raise EntropyError(f"Commit with id '{commit_id}' not found")

        else:
            return self.get_latest_commit()

    def get_latest_commit(self):
        with self.__session_maker() as session:
            commit = (
                session.query(CommitTable)
                .order_by(CommitTable.timestamp.desc())
                .first()
            )
            return commit

    def commit(
        self,
        commit: Commit,
        dirty_keys: Optional[Set[str]] = None,
    ) -> str:
        commit.id = self.__generate_commit_id()
        # TODO: Perhaps create the timestamp here?
        self.stamp_dirty_params_with_commit(commit, dirty_keys)
        commit_table = CommitTable()
        commit_table.id = UUID(commit.id)
        commit_table.timestamp = commit.timestamp
        commit_table.label = commit.label
        commit_table.params = commit.params
        commit_table.tags = commit.tags
        with self.__session_maker() as session:
            session.add(commit_table)
            session.commit()
            return commit.id

    @staticmethod
    def __generate_commit_id() -> str:
        return str(uuid.uuid4())

    def search_commits(
        self, label: Optional[str] = None, key: Optional[str] = None
    ) -> List[Commit]:
        with self.__session_maker() as session:
            commits = session.query(CommitTable)
            if label:
                commits = commits.filter(CommitTable.label == label)
            if key:
                commits = commits.filter(CommitTable.params.contains(key))
            return commits.all()

    def save_temp_commit(self, commit: Commit) -> None:
        with self.__session_maker() as session:
            temp = session.get(TempTable, TEMP_COMMIT_ID)
            if temp:
                temp.timestamp = commit.timestamp
                temp.label = commit.label
                temp.params = commit.params
                temp.tags = commit.tags
                session.commit()
            else:
                temp = TempTable()
                temp.id = TEMP_COMMIT_ID
                temp.timestamp = commit.timestamp
                temp.label = commit.label
                temp.params = commit.params
                temp.tags = commit.tags
                session.add(temp)
                session.commit()

    def load_temp_commit(self) -> Commit:
        with self.__session_maker() as session:
            temp = session.get(TempTable, TEMP_COMMIT_ID)
            if not temp:
                raise EntropyError(
                    "Temp is empty. Use save_temp_commit() before using "
                    "load_temp_commit() "
                )
            else:
                commit = Commit(
                    id=temp.id,
                    timestamp=temp.timestamp,
                    label=temp.label,
                    params=temp.params,
                    tags=temp.tags,
                )
            return commit

    def _db_is_empty(self) -> bool:
        with self.__session_maker() as session:
            result = session.execute(
                text("SELECT sql FROM sqlite_master WHERE type = 'table'")
            )
            return len(result.fetchall()) == 0
